from parsons.databases.database.constants import (
    SMALLINT,
    MEDIUMINT,
    INT,
    BIGINT,
    FLOAT,
    BOOL,
    VARCHAR,
)

from parsons.databases.database.database import DatabaseCreateStatement

import pytest


@pytest.fixture
def dcs():
    db = DatabaseCreateStatement()
    return db


@pytest.fixture
def dcs_bool():
    db = DatabaseCreateStatement()
    db.DO_PARSE_BOOLS = True
    return db


@pytest.mark.parametrize(
    ("int1", "int2", "higher"),
    (
        (SMALLINT, SMALLINT, SMALLINT),
        (SMALLINT, MEDIUMINT, MEDIUMINT),
        (SMALLINT, INT, INT),
        (SMALLINT, BIGINT, BIGINT),
        (MEDIUMINT, SMALLINT, MEDIUMINT),
        (MEDIUMINT, MEDIUMINT, MEDIUMINT),
        (MEDIUMINT, INT, INT),
        (MEDIUMINT, BIGINT, BIGINT),
        (INT, SMALLINT, INT),
        (INT, MEDIUMINT, INT),
        (INT, INT, INT),
        (INT, BIGINT, BIGINT),
        (BIGINT, SMALLINT, BIGINT),
        (BIGINT, MEDIUMINT, BIGINT),
        (BIGINT, INT, BIGINT),
        (BIGINT, BIGINT, BIGINT),
        (None, BIGINT, BIGINT),
        (INT, None, INT),
    ),
)
def test_get_bigger_int(dcs, int1, int2, higher):
    assert dcs.get_bigger_int(int1, int2) == higher


@pytest.mark.parametrize(
    ("val", "is_valid"),
    (
        (10, True),
        (1_0, True),
        (+10, True),
        (+1_0, True),
        (1.2, True),
        (1.0_0, True),
        (1.0, True),
        (1_0.0, True),
        (+1.2, True),
        (+1.0, True),
        (+1.0_0, True),
        (0, True),
        (0.0, True),
        ("10", True),
        ("1_0", False),
        ("+10", True),
        ("+1_0", False),
        ("1.2", True),
        ("1.0_0", False),
        ("1.", True),
        ("1_0.", False),
        ("+1.2", True),
        ("+1.", True),
        ("+1.0_0", False),
        ("0", True),
        ("0.0", True),
        (True, False),
        ("True", False),
        ("a string", False),
        ({}, False),
        ([], False),
        ([], False),
        (None, False),
    ),
)
def test_is_valid_sql_num(dcs, val, is_valid):
    assert dcs.is_valid_sql_num(val) == is_valid


@pytest.mark.parametrize(
    ("val", "cmp_type", "detected_type"),
    (
        (1, None, SMALLINT),
        (1, "", SMALLINT),
        (1, MEDIUMINT, MEDIUMINT),
        (32769, None, MEDIUMINT),
        (32769, BIGINT, BIGINT),
        (2147483648, None, BIGINT),
        (2147483648, FLOAT, FLOAT),
        (5.001, None, FLOAT),
        (5.001, "", FLOAT),
        ("FALSE", VARCHAR, VARCHAR),
        ("word", "", VARCHAR),
        ("word", INT, VARCHAR),
        ("1_2", BIGINT, VARCHAR),
        ("01", FLOAT, VARCHAR),
        ("00001", None, VARCHAR),
        ("word", None, VARCHAR),
        ("1_2", None, VARCHAR),
        ("01", None, VARCHAR),
        ("{}", None, VARCHAR),
    ),
)
def test_detect_data_type(dcs, val, cmp_type, detected_type):
    assert dcs.detect_data_type(val, cmp_type) == detected_type


@pytest.mark.parametrize(
    ("val", "cmp_type", "detected_type"),
    (
        (2, None, SMALLINT),
        (2, "", SMALLINT),
        (1, MEDIUMINT, MEDIUMINT),
        (2, BOOL, SMALLINT),
        (True, None, BOOL),
        (0, None, BOOL),
        (1, None, BOOL),
        (1, BOOL, BOOL),
        ("F", None, BOOL),
        ("FALSE", None, BOOL),
        ("Yes", None, BOOL),
    ),
)
def test_detect_data_type_bools(dcs_bool, val, cmp_type, detected_type):
    assert dcs_bool.detect_data_type(val, cmp_type) == detected_type


@pytest.mark.parametrize(
    ("col", "renamed"),
    (
        ("a", "a"),
        ("A", "a"),
        ("", "_"),
        ("SELECT", "select_"),
        ("two words", "two_words"),
        ("   trailing space   ", "trailing_space"),
        ("1234567890", "x_1234567890"),
        ("0word", "x_0word"),
        # create a really long column name
        # len("asdfghjkla" * 13) == 130
        # len("asdfghjkla" * 10) == 100
        ("asdfghjkla" * 13, "asdfghjkla" * 10),
    ),
)
def test_default_format_column(dcs, col, renamed):
    assert dcs.format_column(col) == renamed


@pytest.mark.parametrize(
    ("cols", "cols_formatted"),
    (
        (
            ["a", "A", "b", "   b   ", "col name", "col_name"],
            ["a", "a_1", "b", "b_3", "col_name", "col_name_5"],
        ),
    ),
)
def test_default_format_columns(dcs, cols, cols_formatted):
    assert dcs.format_columns(cols) == cols_formatted



from suds.client import Client
import logging
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

URI = "https://api.securevan.com/v4/"
SOAP_URI = "https://api.securevan.com/Services/V3/ListService.asmx?WSDL"


class VANConnector(object):
    def __init__(self, api_key=None, auth_name="default", db=None):

        self.api_key = check_env.check("VAN_API_KEY", api_key)

        if db == "MyVoters":
            self.db_code = 0
        elif db in ["MyMembers", "MyCampaign", "EveryAction"]:
            self.db_code = 1
        else:
            raise KeyError(
                "Invalid database type specified. Pick one of:"
                " MyVoters, MyCampaign, MyMembers, EveryAction."
            )

        self.uri = URI
        self.db = db
        self.auth_name = auth_name
        self.pagination_key = "nextPageLink"
        self.auth = (self.auth_name, self.api_key + "|" + str(self.db_code))
        self.api = APIConnector(
            self.uri,
            auth=self.auth,
            data_key="items",
            pagination_key=self.pagination_key,
        )

        # We will not create the SOAP client unless we need to as this triggers checking for
        # valid credentials. As not all API keys are provisioned for SOAP, this keeps it from
        # raising a permission exception when creating the class.
        self._soap_client = None

    @property
    def api_key_profile(self):
        """
        Returns the API key profile with includes permissions and other metadata.
        """

        return self.get_request("apiKeyProfiles")[0]

    @property
    def soap_client(self):

        if not self._soap_client:

            # Create the SOAP client
            soap_auth = {
                "Header": {
                    "DatabaseMode": self.soap_client_db(),
                    "APIKey": self.api_key,
                }
            }
            self._soap_client = Client(SOAP_URI, soapheaders=soap_auth)

        return self._soap_client

    def soap_client_db(self):
        """
        Parse the REST database name to the accepted SOAP format
        """

        if self.db == "MyVoters":
            return "MyVoterFile"
        if self.db == "EveryAction":
            return "MyCampaign"
        else:
            return self.db

    def get_request(self, endpoint, **kwargs):

        r = self.api.get_request(self.uri + endpoint, **kwargs)
        data = self.api.data_parse(r)

        # Paginate
        while isinstance(r, dict) and self.api.next_page_check_url(r):
            if endpoint == "savedLists" and not r["items"]:
                break
            if endpoint == "printedLists" and not r["items"]:
                break
            r = self.api.get_request(r[self.pagination_key], **kwargs)
            data.extend(self.api.data_parse(r))
        return data

    def post_request(self, endpoint, **kwargs):

        return self.api.post_request(endpoint, **kwargs)

    def delete_request(self, endpoint, **kwargs):

        return self.api.delete_request(endpoint, **kwargs)

    def patch_request(self, endpoint, **kwargs):

        return self.api.patch_request(endpoint, **kwargs)

    def put_request(self, endpoint, **kwargs):

        return self.api.put_request(endpoint, **kwargs)



from parsons.databases.database.database import DatabaseCreateStatement
import parsons.databases.mysql.constants as consts

import petl
import logging

logger = logging.getLogger(__name__)


class MySQLCreateTable(DatabaseCreateStatement):
    def __init__(self):
        super().__init__()

        self.VARCHAR_PAD = consts.VARCHAR_PAD
        self.COL_NAME_MAX_LEN = consts.COL_NAME_MAX_LEN
        self.RESERVED_WORDS = []
        self.VARCHAR_STEPS = consts.VARCHAR_STEPS
        self.VARCHAR_MAX = consts.VARCHAR_MAX
        self.SMALLINT_VALUES = consts.SMALLINT_VALUES
        self.MEDIUMINT_VALUES = consts.MEDIUMINT_VALUES
        self.INT_VALUES = consts.INT_VALUES

    # This is for backwards compatability
    def data_type(self, val, current_type):
        return self.detect_data_type(val, current_type)

    # This is for backwards compatability
    def is_valid_integer(self, val):
        return self.is_valid_sql_num(val)

    def evaluate_column(self, column_rows):
        # Generate MySQL data types and widths for a column.

        col_width = 0
        col_type = None

        # Iterate through each row in the column
        for row in column_rows:

            # Get the MySQL data type
            col_type = self.data_type(row, col_type)

            # Calculate width if a varchar
            if col_type == "varchar":
                row_width = len(str(row.encode("utf-8")))

                # Evaluate width vs. current max width
                if row_width > col_width:
                    col_width = row_width

        return col_type, col_width

    def evaluate_table(self, tbl):
        # Generate a dict of MySQL column types and widths for all columns
        # in a table.

        table_map = []

        for col in tbl.columns:
            col_type, col_width = self.evaluate_column(tbl.column_data(col))
            col_map = {"name": col, "type": col_type, "width": col_width}
            table_map.append(col_map)

        return table_map

    def create_statement(self, tbl, table_name, strict_length=True):
        # Generate create statement SQL for a given Parsons table.

        # Validate and rename column names if needed
        tbl.table = petl.setheader(tbl.table, self.columns_convert(tbl.columns))

        # Generate the table map
        table_map = self.evaluate_table(tbl)

        # Generate the column syntax
        column_syntax = []
        for c in table_map:
            if strict_length:
                col_width = int(c["width"] + (self.VARCHAR_PAD * c["width"]))
            else:
                col_width = self.round_longest(c["width"])

            if c["type"] == "varchar":
                column_syntax.append(f"{c['name']} {c['type']}({col_width}) \n")
            else:
                column_syntax.append(f"{c['name']} {c['type']} \n")

        # Generate full statement
        return f"CREATE TABLE {table_name} ( \n {','.join(column_syntax)});"

    # This is for backwards compatability
    def columns_convert(self, columns):
        return self.format_columns(columns, col_prefix="col_")

    @staticmethod
    def round_longest(longest):
        # Find the value that will work best to fit our longest column value
        for step in consts.VARCHAR_STEPS:
            # Make sure we have padding
            if longest < step / 2:
                return step

        return consts.VARCHAR_MAX



import logging
from parsons.ngpvan.events import Events
from parsons.ngpvan.van_connector import VANConnector
from parsons.ngpvan.people import People
from parsons.ngpvan.saved_lists import SavedLists, Folders, ExportJobs
from parsons.ngpvan.activist_codes import ActivistCodes
from parsons.ngpvan.canvass_responses import CanvassResponses
from parsons.ngpvan.survey_questions import SurveyQuestions
from parsons.ngpvan.supporter_groups import SupporterGroups
from parsons.ngpvan.codes import Codes
from parsons.ngpvan.scores import Scores, FileLoadingJobs
from parsons.ngpvan.signups import Signups
from parsons.ngpvan.locations import Locations
from parsons.ngpvan.bulk_import import BulkImport
from parsons.ngpvan.changed_entities import ChangedEntities
from parsons.ngpvan.contact_notes import ContactNotes
from parsons.ngpvan.custom_fields import CustomFields
from parsons.ngpvan.targets import Targets
from parsons.ngpvan.printed_lists import PrintedLists

logger = logging.getLogger(__name__)


class VAN(
    People,
    Events,
    SavedLists,
    PrintedLists,
    Folders,
    ExportJobs,
    ActivistCodes,
    CanvassResponses,
    SurveyQuestions,
    Codes,
    Scores,
    FileLoadingJobs,
    SupporterGroups,
    Signups,
    Locations,
    BulkImport,
    ChangedEntities,
    ContactNotes,
    CustomFields,
    Targets,
):
    """
    Returns the VAN class

    `Args:`
        api_key : str
            A valid api key Not required if ``VAN_API_KEY`` env variable set.
        auth_name: str
            Should not pass this argument
        db: str
            One of ``MyVoters``, ``MyMembers``, ``MyCampaign``, or ``EveryAction``
        uri: str
            Base uri to make api calls.
        raise_for_status: boolean
            Raise excection when encountering a 4XX or 500 error.
    `Returns:`
        VAN object
    """

    def __init__(
        self, api_key=None, auth_name="default", db=None, raise_for_status=True
    ):

        self.connection = VANConnector(api_key=api_key, db=db)
        self.api_key = api_key
        self.db = db

        # The size of each page to return. Currently set to maximum.
        self.page_size = 200



import csv
from io import TextIOWrapper, BytesIO, StringIO
import logging
import sys
import traceback
import time

from parsons.aws.aws_async import (
    get_func_task_path,
    import_and_get_task,
    run as maybe_async_run,
)
from parsons.aws.s3 import S3
from parsons.etl.table import Table
from parsons.utilities.check_env import check

logger = logging.getLogger(__name__)


class DistributeTaskException(Exception):
    pass


class TestStorage:
    def __init__(self):
        self.data = {}

    def put_object(self, bucket, key, object_bytes):
        self.data[key] = object_bytes

    def get_range(self, bucket, key, rangestart, rangeend):
        return self.data[key][rangestart:rangeend]


class S3Storage:
    """
    These methods are pretty specialized, so we keep them
    inside this file rather than s3.py
    """

    def __init__(self, use_env_token=True):
        self.s3 = S3(use_env_token=use_env_token)

    def put_object(self, bucket, key, object_bytes, **kwargs):
        return self.s3.client.put_object(
            Bucket=bucket, Key=key, Body=object_bytes, **kwargs
        )

    def get_range(self, bucket, key, rangestart, rangeend):
        """
        Gets an explicit byte-range of an S3 file
        """
        # bytes is INCLUSIVE for the rangeend parameter, unlike python
        # so e.g. while python returns 2 bytes for data[2:4]
        # Range: bytes=2-4 will return 3!! So we subtract 1
        response = self.s3.client.get_object(
            Bucket=bucket, Key=key, Range="bytes={}-{}".format(rangestart, rangeend - 1)
        )
        return response["Body"].read()


FAKE_STORAGE = TestStorage()
S3_TEMP_KEY_PREFIX = "Parsons_DistributeTask"


def distribute_task_csv(
    csv_bytes_utf8,
    func_to_run,
    bucket,
    header=None,
    func_kwargs=None,
    func_class=None,
    func_class_kwargs=None,
    catch=False,
    group_count=100,
    storage="s3",
    use_s3_env_token=True,
):
    """
    The same as distribute_task, but instead of a table, the
    first argument is bytes of a csv encoded into utf8.
    This function is used by distribute_task() which you should use instead.
    """
    global FAKE_STORAGE
    func_name = get_func_task_path(func_to_run, func_class)
    row_chunks = csv_bytes_utf8.split(b"\n")
    cursor = 0
    row_ranges = []
    # gather start/end bytes for each row
    for rowc in row_chunks:
        rng = [cursor]
        cursor = cursor + len(rowc) + 1  # +1 is the \n character
        rng.append(cursor)
        row_ranges.append(rng)

    # group the rows and get start/end bytes for each group
    group_ranges = []
    # table csv writer appends a terminal \r\n, so we do len-1
    for grpstep in range(0, len(row_ranges) - 1, group_count):
        end = min(len(row_ranges) - 1, grpstep + group_count - 1)
        group_ranges.append((row_ranges[grpstep][0], row_ranges[end][1]))

    # upload data
    filename = hash(time.time())
    storagekey = f"{S3_TEMP_KEY_PREFIX}/{filename}.csv"
    groupcount = len(group_ranges)
    logger.debug(f"distribute_task_csv storagekey {storagekey} w/ {groupcount} groups")

    response = None
    if storage == "s3":
        response = S3Storage(use_env_token=use_s3_env_token).put_object(
            bucket, storagekey, csv_bytes_utf8
        )
    else:
        response = FAKE_STORAGE.put_object(bucket, storagekey, csv_bytes_utf8)

    # start processes
    results = [
        maybe_async_run(
            process_task_portion,
            [
                bucket,
                storagekey,
                grp[0],
                grp[1],
                func_name,
                header,
                storage,
                func_kwargs,
                catch,
                func_class_kwargs,
                use_s3_env_token,
            ],
            # if we are using local storage, then it must be run locally, as well
            # (good for testing/debugging)
            remote_aws_lambda_function_name="FORCE_LOCAL"
            if storage == "local"
            else None,
        )
        for grp in group_ranges
    ]
    return {
        "DEBUG_ONLY": "results may vary depending on context/platform",
        "results": results,
        "put_response": response,
    }


def distribute_task(
    table,
    func_to_run,
    bucket=None,
    func_kwargs=None,
    func_class=None,
    func_class_kwargs=None,
    catch=False,
    group_count=100,
    storage="s3",
    use_s3_env_token=True,
):
    """
    Distribute processing rows in a table across multiple AWS Lambda invocations.

    `Args:`
        table: Parsons Table
           Table of data you wish to distribute processing across Lambda invocations
           of `func_to_run` argument.
        func_to_run: function
           The function you want to run whose
           first argument will be a subset of table
        bucket: str
           The bucket name to use for s3 upload to process the whole table
           Not required if you set environment variable ``S3_TEMP_BUCKET``
        func_kwargs: dict
           If the function has other arguments to pass along with `table`
           then provide them as a dict here. They must all be JSON-able.
        func_class: class
           If the function is a classmethod or function on a class,
           then pass the pure class here.
           E.g. If you passed `ActionKit.bulk_upload_table`,
           then you would pass `ActionKit` here.
        func_class_kwargs: dict
           If it is a class function, and the class must be instantiated,
           then pass the kwargs to instantiate the class here.
           E.g. If you passed `ActionKit.bulk_upload_table` as the function,
           then you would pass {'domain': ..., 'username': ... etc} here.
           This must all be JSON-able data.
        catch: bool
           Lambda will retry running an event three times if there's an
           exception -- if you want to prevent this, set `catch=True`
           and then it will catch any errors and stop retries.
           The error will be in CloudWatch logs with string "Distribute Error"
           This might be important if row-actions are not idempotent and your
           own function might fail causing repeats.
        group_count: int
           Set this to how many rows to process with each Lambda invocation (Default: 100)
        storage: str
           Debugging option: Defaults to "s3". To test distribution locally without s3,
           set to "local".
        use_s3_env_token: str
           If storage is set to "s3", sets the use_env_token parameter on the S3 storage.
    `Returns:`
        Debug information -- do not rely on the output, as it will change
        depending on how this method is invoked.
    """
    if storage not in ("s3", "local"):
        raise DistributeTaskException("storage argument must be s3 or local")
    bucket = check("S3_TEMP_BUCKET", bucket)
    csvdata = StringIO()
    outcsv = csv.writer(csvdata)
    outcsv.writerows(table.table.data())
    return distribute_task_csv(
        csvdata.getvalue().encode("utf-8-sig"),
        func_to_run,
        bucket,
        header=table.columns,
        func_kwargs=func_kwargs,
        func_class=func_class,
        func_class_kwargs=func_class_kwargs,
        catch=catch,
        group_count=group_count,
        storage=storage,
        use_s3_env_token=use_s3_env_token,
    )


def process_task_portion(
    bucket,
    storagekey,
    rangestart,
    rangeend,
    func_name,
    header,
    storage="s3",
    func_kwargs=None,
    catch=False,
    func_class_kwargs=None,
    use_s3_env_token=True,
):
    global FAKE_STORAGE

    logger.debug(
        f"process_task_portion func_name {func_name}, "
        f"storagekey {storagekey}, byterange {rangestart}-{rangeend}"
    )
    func = import_and_get_task(func_name, func_class_kwargs)
    if storage == "s3":
        filedata = S3Storage(use_env_token=use_s3_env_token).get_range(
            bucket, storagekey, rangestart, rangeend
        )
    else:
        filedata = FAKE_STORAGE.get_range(bucket, storagekey, rangestart, rangeend)

    lines = list(csv.reader(TextIOWrapper(BytesIO(filedata), encoding="utf-8-sig")))
    table = Table([header] + lines)
    if catch:
        try:
            func(table, **func_kwargs)
        except Exception:
            # In Lambda you can search for '"Distribute Error"' in the logs
            type_, value_, traceback_ = sys.exc_info()
            err_traceback_str = "\n".join(
                traceback.format_exception(type_, value_, traceback_)
            )
            return {
                "Exception": "Distribute Error",
                "error": err_traceback_str,
                "rangestart": rangestart,
                "rangeend": rangeend,
                "func_name": func_name,
                "bucket": bucket,
                "storagekey": storagekey,
            }
    else:
        func(table, **func_kwargs)



ENV_PARAMETERS = {"SISENSE_SITE_NAME": "my_site_name", "SISENSE_API_KEY": "my_api_key"}

TEST_PUBLISH_SHARED_DASHBOARD = {
    "url": "https://www.periscopedata.com/api/embedded_dashboard?data=%7B%22dashboard%22%3A7863%2C%22embed%22%3A%22v2%22%2C%22filters%22%3A%5B%7B%22name%22%3A%22Filter1%22%2C%22value%22%3A%22value1%22%7D%2C%7B%22name%22%3A%22Filter2%22%2C%22value%22%3A%221234%22%7D%5D%7D&signature=adcb671e8e24572464c31e8f9ffc5f638ab302a0b673f72554d3cff96a692740"  # noqa: E501
}

TEST_LIST_SHARED_DASHBOARDS = [
    "https://app.periscopedata.com/shared/abc",
    "https://app.periscopedata.com/shared/def",
]

TEST_DELETE_SHARED_DASHBOARD = {"success": True}



from parsons import Freshdesk
import unittest
import requests_mock
from test.test_freshdesk import expected_json

DOMAIN = "myorg"
API_KEY = "mykey"


class TestFreshdesk(unittest.TestCase):
    def setUp(self):

        self.fd = Freshdesk(DOMAIN, API_KEY)

    @requests_mock.Mocker()
    def test_get_agents(self, m):

        # Test that agents are returned correctly.
        m.get(self.fd.uri + "agents", json=expected_json.test_agent)
        self.fd.get_agents()

    @requests_mock.Mocker()
    def test_get_tickets(self, m):

        # Test that tickets are returned correctly.
        m.get(self.fd.uri + "tickets", json=expected_json.test_ticket)
        self.fd.get_tickets()

    @requests_mock.Mocker()
    def test_get_companies(self, m):

        # Test that tickets are returned correctly.
        m.get(self.fd.uri + "companies", json=expected_json.test_company)
        self.fd.get_companies()

    @requests_mock.Mocker()
    def test_get_contacts(self, m):

        # Test that tickets are returned correctly.
        m.get(self.fd.uri + "contacts", json=expected_json.test_contact)
        self.fd.get_contacts()



import unittest
from unittest.mock import patch
from parsons.databases.redshift import Redshift
from parsons.databases.mysql import MySQL
from parsons.databases.postgres import Postgres
from parsons.google.google_bigquery import GoogleBigQuery
from parsons.databases.discover_database import discover_database


class TestDiscoverDatabase(unittest.TestCase):
    @patch.object(GoogleBigQuery, "__init__", return_value=None)
    @patch.object(Postgres, "__init__", return_value=None)
    @patch.object(MySQL, "__init__", return_value=None)
    @patch.object(Redshift, "__init__", return_value=None)
    @patch("os.getenv")
    def test_no_database_detected(self, mock_getenv, *_):
        mock_getenv.return_value = None
        with self.assertRaises(EnvironmentError):
            discover_database()

    @patch.object(GoogleBigQuery, "__init__", return_value=None)
    @patch.object(Postgres, "__init__", return_value=None)
    @patch.object(MySQL, "__init__", return_value=None)
    @patch.object(Redshift, "__init__", return_value=None)
    @patch("os.getenv")
    def test_single_database_detected(self, mock_getenv, *_):
        mock_getenv.side_effect = (
            lambda var: "password" if var == "REDSHIFT_PASSWORD" else None
        )
        self.assertIsInstance(discover_database(), Redshift)

    @patch.object(GoogleBigQuery, "__init__", return_value=None)
    @patch.object(Postgres, "__init__", return_value=None)
    @patch.object(MySQL, "__init__", return_value=None)
    @patch.object(Redshift, "__init__", return_value=None)
    @patch("os.getenv")
    def test_single_database_detected_with_other_default(self, mock_getenv, *_):
        mock_getenv.side_effect = (
            lambda var: "password" if var == "REDSHIFT_PASSWORD" else None
        )
        self.assertIsInstance(discover_database(default_connector=Postgres), Redshift)

    @patch.object(GoogleBigQuery, "__init__", return_value=None)
    @patch.object(Postgres, "__init__", return_value=None)
    @patch.object(MySQL, "__init__", return_value=None)
    @patch.object(Redshift, "__init__", return_value=None)
    @patch("os.getenv")
    def test_single_database_detected_with_other_default_list(self, mock_getenv, *_):
        mock_getenv.side_effect = (
            lambda var: "password" if var == "REDSHIFT_PASSWORD" else None
        )
        self.assertIsInstance(
            discover_database(default_connector=[Postgres, MySQL]), Redshift
        )

    @patch.object(GoogleBigQuery, "__init__", return_value=None)
    @patch.object(Postgres, "__init__", return_value=None)
    @patch.object(MySQL, "__init__", return_value=None)
    @patch.object(Redshift, "__init__", return_value=None)
    @patch("os.getenv")
    def test_multiple_databases_no_default(self, mock_getenv, *_):
        mock_getenv.return_value = "password"
        with self.assertRaises(EnvironmentError):
            discover_database()

    @patch.object(GoogleBigQuery, "__init__", return_value=None)
    @patch.object(Postgres, "__init__", return_value=None)
    @patch.object(MySQL, "__init__", return_value=None)
    @patch.object(Redshift, "__init__", return_value=None)
    @patch("os.getenv")
    def test_multiple_databases_with_default(self, mock_getenv, *_):
        mock_getenv.return_value = "password"
        self.assertIsInstance(discover_database(default_connector=Redshift), Redshift)

    @patch.object(GoogleBigQuery, "__init__", return_value=None)
    @patch.object(Postgres, "__init__", return_value=None)
    @patch.object(MySQL, "__init__", return_value=None)
    @patch.object(Redshift, "__init__", return_value=None)
    @patch("os.getenv")
    def test_multiple_databases_with_default_list(self, mock_getenv, *_):
        mock_getenv.return_value = "password"
        self.assertIsInstance(
            discover_database(default_connector=[MySQL, Redshift]), MySQL
        )

    @patch.object(GoogleBigQuery, "__init__", return_value=None)
    @patch.object(Postgres, "__init__", return_value=None)
    @patch.object(MySQL, "__init__", return_value=None)
    @patch.object(Redshift, "__init__", return_value=None)
    @patch("os.getenv")
    def test_multiple_databases_invalid_default(self, mock_getenv, *_):
        mock_getenv.side_effect = (
            lambda var: "password"
            if var == "REDSHIFT_PASSWORD" or var == "MYSQL_PASSWORD"
            else None
        )
        with self.assertRaises(EnvironmentError):
            discover_database(default_connector=Postgres)

    @patch.object(GoogleBigQuery, "__init__", return_value=None)
    @patch.object(Postgres, "__init__", return_value=None)
    @patch.object(MySQL, "__init__", return_value=None)
    @patch.object(Redshift, "__init__", return_value=None)
    @patch("os.getenv")
    def test_multiple_databases_invalid_default_list(self, mock_getenv, *_):
        mock_getenv.side_effect = (
            lambda var: "password"
            if var == "REDSHIFT_PASSWORD" or var == "MYSQL_PASSWORD"
            else None
        )
        with self.assertRaises(EnvironmentError):
            discover_database(default_connector=[Postgres, GoogleBigQuery])


if __name__ == "__main__":
    unittest.main()



from oauth2client.service_account import ServiceAccountCredentials
from parsons.etl.table import Table
from parsons.google.utitities import setup_google_application_credentials
import httplib2
import json
import os


class GoogleAdmin(object):
    """
    A connector for Google Admin.


    `Args:`
        app_creds: str
            A credentials json string or a path to a json file. Not required if
            ``GOOGLE_APPLICATION_CREDENTIALS`` env variable set.
        sub: str
            An email address that this service account will act on behalf of (via domain-wide
            delegation)
    `Returns:`
        GoogleAdmin Class
    """

    def __init__(self, app_creds=None, sub=None):
        setup_google_application_credentials(app_creds)

        self.client = (
            ServiceAccountCredentials.from_json_keyfile_name(
                os.environ["GOOGLE_APPLICATION_CREDENTIALS"],
                ["https://www.googleapis.com/auth/admin.directory.group"],
            )
            .create_delegated(sub)
            .authorize(httplib2.Http())
        )

    def _paginate_request(self, endpoint, collection, params=None):
        # Build query params
        param_arr = []
        param_str = ""
        if params:
            for key, value in params.items():
                param_arr.append(key + "=" + value)
            param_str = "?" + "&".join(param_arr)

        # Make API call
        req_url = "https://admin.googleapis.com/admin/directory/v1/" + endpoint

        # Return type from Google Admin is a tuple of length 2. Extract desired result from 2nd item
        # in tuple and convert to json
        res = json.loads(
            self.client.request(req_url + param_str, "GET")[1].decode("utf-8")
        )

        # Paginate
        ret = []
        if collection in res:
            ret = res[collection]

            while "nextPageToken" in res:
                if param_arr[-1][0:10] != "pageToken=":
                    param_arr.append("pageToken=" + res["nextPageToken"])
                else:
                    param_arr[-1] = "pageToken=" + res["nextPageToken"]
                res = json.loads(
                    self.client.request(req_url + "?" + "&".join(param_arr), "GET")[
                        1
                    ].decode("utf-8")
                )
                ret += res[collection]

        return Table(ret)

    def get_aliases(self, group_key, params=None):
        """
        Get aliases for a group. `Google Admin API Documentation <https://developers.google.com/\
        admin-sdk/directory/reference/rest/v1/groups.aliases/list>`_

        `Args:`
            group_key: str
                The Google group id
            params: dict
                A dictionary of fields for the GET request
        `Returns:`
            Table Class
        """
        return self._paginate_request(
            "groups/" + group_key + "/aliases", "aliases", params
        )

    def get_all_group_members(self, group_key, params=None):
        """
        Get all members in a group. `Google Admin API Documentation <https://developers.google.com/\
        admin-sdk/directory/v1/guides/manage-group-members#get_all_members>`_

        `Args:`
            group_key: str
                The Google group id
            params: dict
                A dictionary of fields for the GET request
        `Returns:`
            Table Class
        """
        return self._paginate_request(
            "groups/" + group_key + "/members", "members", params
        )

    def get_all_groups(self, params=None):
        """
        Get all groups in a domain or account. `Google Admin API Documentation <https://developers.\
        google.com/admin-sdk/directory/v1/guides/manage-groups#get_all_domain_groups>`_
        `Args:`
            params: dict
                A dictionary of fields for the GET request.
        `Returns:`
            Table Class
        """
        return self._paginate_request("groups", "groups", params)



import unittest
import os
import requests_mock
import unittest.mock as mock
from parsons import VAN, Table
from test.utils import assert_matching_tables


class TestNGPVAN(unittest.TestCase):
    def setUp(self):

        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters", raise_for_status=False)

    @requests_mock.Mocker()
    def test_get_changed_entity_resources(self, m):

        json = ["ActivistCodes", "ContactHistory", "Contacts", "ContactsActivistCodes"]
        m.get(self.van.connection.uri + "changedEntityExportJobs/resources", json=json)
        self.assertEqual(json, self.van.get_changed_entity_resources())

    @requests_mock.Mocker()
    def test_get_changed_entity_resource_fields(self, m):

        json = [
            {
                "fieldName": "ActivistCodeID",
                "fieldType": "N",
                "maxTextboxCharacters": None,
                "isCoreField": True,
                "availableValues": None,
            },
            {
                "fieldName": "ActivistCodeType",
                "fieldType": "T",
                "maxTextboxCharacters": 20,
                "isCoreField": True,
                "availableValues": None,
            },
            {
                "fieldName": "Campaign",
                "fieldType": "T",
                "maxTextboxCharacters": 150,
                "isCoreField": True,
                "availableValues": None,
            },
        ]

        m.get(
            self.van.connection.uri + "changedEntityExportJobs/fields/ActivistCodes",
            json=json,
        )
        assert_matching_tables(
            Table(json), self.van.get_changed_entity_resource_fields("ActivistCodes")
        )

    @requests_mock.Mocker()
    def test_get_changed_entities(self, m):

        json = {
            "dateChangedFrom": "2021-10-10T00:00:00-04:00",
            "dateChangedTo": "2021-10-11T00:00:00-04:00",
            "files": [],
            "message": "Created export job",
            "code": None,
            "exportedRecordCount": 0,
            "exportJobId": 2170181229,
            "jobStatus": "Pending",
        }

        json2 = {
            "dateChangedFrom": "2021-10-10T00:00:00-04:00",
            "dateChangedTo": "2021-10-11T00:00:00-04:00",
            "files": [
                {
                    "downloadUrl": "https://box.com/file.csv",
                    "dateExpired": "2021-11-03T15:27:01.8687339-04:00",
                }
            ],
            "message": "Finished processing export job",
            "code": None,
            "exportedRecordCount": 6110,
            "exportJobId": 2170181229,
            "jobStatus": "Complete",
        }

        tbl = Table([{"a": 1, "b": 2}])

        m.post(self.van.connection.uri + "changedEntityExportJobs", json=json)
        m.get(
            self.van.connection.uri + "changedEntityExportJobs/2170181229", json=json2
        )

        Table.from_csv = mock.MagicMock()
        Table.from_csv.return_value = tbl

        out_tbl = self.van.get_changed_entities("ContactHistory", "2021-10-10")

        assert_matching_tables(out_tbl, tbl)



import unittest
import os
import requests_mock
from parsons import VAN
from test.utils import validate_list

signup_status = [
    {"statusId": 5, "name": "Cancelled"},
    {"statusId": 2, "name": "Completed"},
    {"statusId": 11, "name": "Confirmed"},
    {"statusId": 4, "name": "Invited"},
    {"statusId": 18, "name": "Left Msg"},
    {"statusId": 6, "name": "No Show"},
    {"statusId": 1, "name": "Scheduled"},
    {"statusId": 30, "name": "Sched-Web"},
    {"statusId": 15, "name": "Walk In"},
]

signup = {
    "eventSignupId": 14285,
    "person": {
        "vanId": 100349920,
        "firstName": "Helen",
        "middleName": None,
        "lastName": "Maddix",
        "contactOrganizationCommonName": None,
        "contactOrganizationOfficialName": None,
        "contactModeId": 1,
        "email": None,
        "phone": None,
    },
    "event": {"eventId": 750001004, "name": "Canvass 01", "shortName": "Can01"},
    "shift": {"eventShiftId": 19076, "name": "Default Shift"},
    "role": {"roleId": 263920, "name": "Leader"},
    "status": {"statusId": 11, "name": "Confirmed"},
    "location": {
        "locationId": 3,
        "name": "First Presbyterian Church",
        "displayName": "First Presbyterian Church, 340 5th Ave S Saint Cloud, MN 56301 ",
    },
    "startTimeOverride": "2018-12-31T08:00:00-05:00",
    "endTimeOverride": "2018-12-31T10:00:00-05:00",
    "printedLists": None,
    "minivanExports": None,
    "supporterGroupId": None,
    "isOfflineSignup": True,
}

signup_expected = [
    "eventSignupId",
    "startTimeOverride",
    "endTimeOverride",
    "printedLists",
    "minivanExports",
    "supporterGroupId",
    "isOfflineSignup",
    "contactModeId",
    "contactOrganizationCommonName",
    "contactOrganizationOfficialName",
    "email",
    "firstName",
    "lastName",
    "middleName",
    "phone",
    "vanId",
    "status_name",
    "status_statusId",
    "event_eventId",
    "event_name",
    "event_shortName",
    "shift_eventShiftId",
    "shift_name",
    "role_name",
    "role_roleId",
    "location_displayName",
    "location_locationId",
    "location_name",
]

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestSignups(unittest.TestCase):
    def setUp(self):

        self.van = VAN(
            os.environ["VAN_API_KEY"], db="EveryAction", raise_for_status=False
        )

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_signup_statuses(self, m):

        m.get(self.van.connection.uri + "signups/statuses", json=signup_status)

        # Test events lookup
        self.assertTrue(
            validate_list(
                ["statusId", "name"], self.van.get_signups_statuses(event_id=750000849)
            )
        )

        # Test event type lookup
        self.assertTrue(
            validate_list(
                ["statusId", "name"],
                self.van.get_signups_statuses(event_type_id=750000849),
            )
        )

    @requests_mock.Mocker()
    def test_get_signups(self, m):

        json = {"items": [signup], "nextPageLink": None, "count": 1}

        m.get(self.van.connection.uri + "signups", json=json)

        self.assertTrue(
            validate_list(
                signup_expected, self.van.get_event_signups(event_id=750001004)
            )
        )

        self.assertTrue(
            validate_list(signup_expected, self.van.get_person_signups(vanid=750000849))
        )

    @requests_mock.Mocker()
    def test_get_signup(self, m):

        event_signup_id = 14285

        m.get(
            self.van.connection.uri + f"signups/{event_signup_id}".format(), json=signup
        )

        self.assertEqual(signup, self.van.get_signup(event_signup_id))

    @requests_mock.Mocker()
    def test_create_signup(self, m):

        m.post(self.van.connection.uri + "signups", json=14285, status_code=201)

        self.assertEqual(
            self.van.create_signup(100349920, 750001004, 19076, 263920, 11, 3), 14285
        )

    @requests_mock.Mocker()
    def test_update_signup(self, m):

        # This is two part. It makes a call to get the object and then it updates it

        event_signup_id = 14285

        # Get object route
        m.get(self.van.connection.uri + f"signups/{event_signup_id}", json=signup)

        # Update object
        m.put(self.van.connection.uri + f"signups/{event_signup_id}", status_code=204)

        self.van.update_signup(event_signup_id, status_id=6)



import json
import logging
import time
from typing import Any, Dict, Optional, Tuple, cast
from urllib.parse import parse_qs, urlparse

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)


class NationBuilder:
    """
    Instantiate the NationBuilder class

    `Args:`
        slug: str
            The Nation Builder slug Not required if ``NB_SLUG`` env variable set. The slug is the
            nation slug of the nation from which your application is requesting approval to retrieve
            data via the NationBuilder API. For example, your application's user could provide this
            slug via a text field in your application.
        access_token: str
            The Nation Builder access_token Not required if ``NB_ACCESS_TOKEN`` env variable set.
    """

    def __init__(
        self, slug: Optional[str] = None, access_token: Optional[str] = None
    ) -> None:
        slug = check_env.check("NB_SLUG", slug)
        token = check_env.check("NB_ACCESS_TOKEN", access_token)

        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        headers.update(NationBuilder.get_auth_headers(token))

        self.client = APIConnector(NationBuilder.get_uri(slug), headers=headers)

    @classmethod
    def get_uri(cls, slug: Optional[str]) -> str:
        if slug is None:
            raise ValueError("slug can't None")

        if not isinstance(slug, str):
            raise ValueError("slug must be an str")

        if len(slug.strip()) == 0:
            raise ValueError("slug can't be an empty str")

        return f"https://{slug}.nationbuilder.com/api/v1"

    @classmethod
    def get_auth_headers(cls, access_token: Optional[str]) -> Dict[str, str]:
        if access_token is None:
            raise ValueError("access_token can't None")

        if not isinstance(access_token, str):
            raise ValueError("access_token must be an str")

        if len(access_token.strip()) == 0:
            raise ValueError("access_token can't be an empty str")

        return {"authorization": f"Bearer {access_token}"}

    @classmethod
    def parse_next_params(cls, next_value: str) -> Tuple[str, str]:
        next_params = parse_qs(urlparse(next_value).query)

        if "__nonce" not in next_params:
            raise ValueError("__nonce param not found")

        if "__token" not in next_params:
            raise ValueError("__token param not found")

        nonce = next_params["__nonce"][0]
        token = next_params["__token"][0]

        return nonce, token

    @classmethod
    def make_next_url(cls, original_url: str, nonce: str, token: str) -> str:
        return f"{original_url}?limit=100&__nonce={nonce}&__token={token}"

    def get_people(self) -> Table:
        """
        `Returns:`
            A Table of all people stored in Nation Builder.
        """
        data = []
        original_url = "people"

        url = f"{original_url}"

        while True:
            try:
                logging.debug("sending request %s" % url)
                response = self.client.get_request(url)

                res = response.get("results", None)

                if res is None:
                    break

                logging.debug("response got %s records" % len(res))

                data.extend(res)

                if response.get("next", None):
                    nonce, token = NationBuilder.parse_next_params(response["next"])
                    url = NationBuilder.make_next_url(original_url, nonce, token)
                else:
                    break
            except Exception as error:
                logging.error("error requesting data from Nation Builder: %s" % error)

                wait_time = 30
                logging.info("waiting %d seconds before retrying" % wait_time)
                time.sleep(wait_time)

        return Table(data)

    def update_person(self, person_id: str, person: Dict[str, Any]) -> Dict[str, Any]:
        """
        This method updates a person with the provided id to have the provided data. It returns a
        full representation of the updated person.

        `Args:`
            person_id: str
                Nation Builder person id.
            data: dict
                Nation builder person object.
                For example {"email": "user@example.com", "tags": ["foo", "bar"]}
                Docs: https://nationbuilder.com/people_api
        `Returns:`
            A person object with the updated data.
        """
        if person_id is None:
            raise ValueError("person_id can't None")

        if not isinstance(person_id, str):
            raise ValueError("person_id must be a str")

        if len(person_id.strip()) == 0:
            raise ValueError("person_id can't be an empty str")

        if not isinstance(person, dict):
            raise ValueError("person must be a dict")

        url = f"people/{person_id}"
        response = self.client.put_request(url, data=json.dumps({"person": person}))
        response = cast(Dict[str, Any], response)

        return response

    def upsert_person(
        self, person: Dict[str, Any]
    ) -> Tuple[bool, Optional[Dict[str, Any]]]:
        """
        Updates a matched person or creates a new one if the person doesn't exist.

        This method attempts to match the input person resource to a person already in the
        nation. If a match is found, the matched person is updated. If a match is not found, a new
        person is created. Matches are found by including one of the following IDs in the request:

            - civicrm_id
            - county_file_id
            - dw_id
            - external_id
            - email
            - facebook_username
            - ngp_id
            - salesforce_id
            - twitter_login
            - van_id

        `Args:`
            data: dict
                Nation builder person object.
                For example {"email": "user@example.com", "tags": ["foo", "bar"]}
                Docs: https://nationbuilder.com/people_api
        `Returns:`
            A tuple of `created` and `person` object with the updated data. If the request fails
            the method will return a tuple of `False` and `None`.
        """

        _required_keys = [
            "civicrm_id",
            "county_file_id",
            "dw_id",
            "external_id",
            "email",
            "facebook_username",
            "ngp_id",
            "salesforce_id",
            "twitter_login",
            "van_id",
        ]

        if not isinstance(person, dict):
            raise ValueError("person must be a dict")

        has_required_key = any(x in person for x in _required_keys)

        if not has_required_key:
            _keys = ", ".join(_required_keys)
            raise ValueError(f"person dict must contain at least one key of {_keys}")

        url = "people/push"
        response = self.client.request(url, "PUT", data=json.dumps({"person": person}))

        self.client.validate_response(response)

        if response.status_code == 200:
            if self.client.json_check(response):
                return (False, response.json())

        if response.status_code == 201:
            if self.client.json_check(response):
                return (True, response.json())

        return (False, None)



from requests import request
import time
from parsons.etl import Table
from parsons.utilities import check_env
import logging

logger = logging.getLogger(__name__)

CT_URI = "https://api.crowdtangle.com/"
PAGE_SIZE = 100
REQUEST_SLEEP = 10  # CT has a rather agressive 6 requests per minute rate limit.


class CrowdTangle(object):
    """
    Instantiate CrowdTangle Class

    `Args:`
        api_key: str
            A valid CrowdTangle API key. Not required if ``CROWDTANGLE_API_KEY`` env
            variable set.
    `Returns:`
        CrowdTangle Class
    """

    def __init__(self, api_key=None):

        self.api_key = check_env.check("CT_API_KEY", api_key)
        self.uri = CT_URI

    def _base_request(self, endpoint, req_type="GET", args=None):

        url = f"{self.uri}/{endpoint}"
        base_args = {"token": self.api_key, "count": PAGE_SIZE}

        # Add any args passed through to the base args
        if args is not None:
            base_args.update(args)

        r = request(req_type, url, params=base_args).json()
        json = r["result"]
        keys = list(json.keys())
        data = json[keys[0]]

        while "nextPage" in list(json["pagination"].keys()):
            logger.info(f"Retrieving {PAGE_SIZE} rows.")
            time.sleep(REQUEST_SLEEP)
            next_url = json["pagination"]["nextPage"]
            r = request(req_type, next_url).json()
            json = r["result"]
            data.extend(json[keys[0]])

        logger.info(f"Retrieved {len(data)} rows.")

        return data

    def _base_unpack(self, ParsonsTable):

        logger.debug("Working to unpack the Parsons Table...")
        logger.debug(f"Starting with {len(ParsonsTable.columns)} columns...")
        sample = ParsonsTable[0]

        col_dict = {}
        for x in ParsonsTable.columns:
            col_dict[x] = str(type(sample[x]))

        for col in col_dict:
            if col_dict[col] == "<class 'dict'>":
                ParsonsTable.unpack_dict(col)
            elif col_dict[x] == "<class 'list'>":
                ParsonsTable.unpack_list(col)

        logger.info(f"There are now {len(ParsonsTable.columns)} columns...")
        return ParsonsTable

    def _unpack(self, ParsonsTable):

        if ParsonsTable.num_rows == 0:
            return None

        first_count = len(ParsonsTable.columns)
        second_count = first_count + 1

        while second_count > first_count:
            first_count = len(ParsonsTable.columns)
            self._base_unpack(ParsonsTable)
            second_count = len(ParsonsTable.columns)

        return ParsonsTable

    def _list_to_string(self, list_arg):

        if list_arg:
            return ",".join(list_arg)
        else:
            return None

    def get_posts(
        self,
        start_date=None,
        end_date=None,
        language=None,
        list_ids=None,
        min_interations=None,
        search_term=None,
        types=None,
    ):
        """
        Return a set of posts for the given parameters.

        See the `API documentation <https://github.com/CrowdTangle/API/wiki/Posts>`_
        for more information.

        .. warning::
          Rate limit is 2 calls / minute.

        `Args:`
            start_date: str
                Filter to the earliest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss`` or ``yyyy-mm-dd``).
            end_date: str
                Filter to the latest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss`` or ``yyyy-mm-dd``).
            language: str
                Filter to 2-character Locale code. Some languages require more
                than two characters: Chinese (Simplified) is zh-CN and
                Chinese (Traditional) is zh-TW.
            list_ids: list
                Filter to the ids of lists or saved searches to retrieve.
            min_interactions: int
                Filter to posts with total interactions above this threshold.
            search_team: str
                Returns only posts that match this search term. For multiple terms, separate
                with commas for OR, use quotes for phrases.
            types: list
                Filter to post types including:
                * ``episode``
                * ``extra_clip``
                * ``link``
                * ``live_video``
                * ``live_video_complete``
                * ``live_video_scheduled``
                * ``native_video``
                * ``photo``
                * ``status``
                * ``trailer``
                * ``tweet``
                * ``vimeo``
                * ``vine``
                * ``youtube``

                If you want all live videos (whether currently or formerly live),
                pass include both ``live_video`` and ``live_video_complete``
                parameters.

                The ``video`` type does not mean all videos, it refers to videos
                that are not ``native_video``, ``youtube`` or ``vine``.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        args = {
            "startDate": start_date,
            "endDate": end_date,
            "language": language,
            "listIds": self._list_to_string(list_ids),
            "minInteractions": min_interations,
            "searchTerm": search_term,
            "types": types,
        }

        logger.info("Retrieving posts.")
        pt = Table(self._base_request("posts", args=args))
        logger.info(f"Retrieved {pt.num_rows} posts.")
        self._unpack(pt)
        return pt

    def get_leaderboard(
        self, start_date=None, end_date=None, list_ids=None, account_ids=None
    ):
        """
        Return leaderboard data.

        See the `API documentation <https://github.com/CrowdTangle/API/wiki/Leaderboard>`_
        for more information.

        .. warning::
          Rate limit is 6 calls / minute.

        `Args:`
            start_date: str
                Filter to the earliest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss`` or ``yyyy-mm-dd``).
            end_date: str
                Filter to the latest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss`` or ``yyyy-mm-dd``).
            list_ids: list
                Filter to the ids of lists or saved searches to retrieve.
            account_ids: list
                A list of CrowdTangle accountIds to retrieve leaderboard data for.
                This and ``list_id`` are mutually exclusive; if both are sent, the
                ``account_ids`` value will be used.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        args = {
            "startDate": start_date,
            "endDate": end_date,
            "listIds": self._list_to_string(list_ids),
            "accountIds": self._list_to_string(account_ids),
        }

        pt = Table(self._base_request("leaderboard", args=args))
        logger.info(f"Retrieved {pt.num_rows} records from the leaderbooard.")
        self._unpack(pt)
        return pt

    def get_links(
        self, link, start_date=None, end_date=None, include_summary=None, platforms=None
    ):
        """
        Return up to 100 posts based on a specific link. It is strongly recommended to
        use the ``start_date`` parameter to limit queries to relevant dates.

        See the `API documentation <https://github.com/CrowdTangle/API/wiki/Links>`_
        for more information.

        .. warning::
          Rate limit is 2 calls / minute.

        `Args:`
            link: str
                The link to filter posts to.
            start_date: str
                Filter to the earliest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss`` or ``yyyy-mm-dd``).
            end_date: str
                Filter to the latest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss`` or ``yyyy-mm-dd``).
            include_summary: boolean
                Adds a ``summary`` column with account statistics for each platform
                that has posted this link. It will look beyond the count
                requested to summarize across the time searched.
                Requires a value for ``start_date``.
            platforms: list
                Filter by platforms

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        args = {
            "link": link,
            "startDate": start_date,
            "endDate": end_date,
            "includeSummary": str(include_summary),
            "platforms": self._list_to_string(platforms),
        }

        logger.info("Retrieving posts based on link.")
        pt = Table(self._base_request("links", args=args))
        logger.info(f"Retrieved {pt.num_rows} links.")
        self._unpack(pt)
        return pt



import unittest
import os
import requests_mock
from parsons import VAN
from test.utils import assert_matching_tables
from requests.exceptions import HTTPError

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestCodes(unittest.TestCase):
    def setUp(self):

        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters")

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_codes(self, m):

        json = {
            "items": [
                {
                    "codeId": 1004916,
                    "parentCodeId": None,
                    "name": "Data Entry",
                    "description": "for test.",
                    "codePath": "Data Entry",
                    "createdByName": "",
                    "dateCreated": "2018-07-13T15:16:00Z",
                    "supportedEntities": None,
                    "codeType": "Tag",
                    "campaign": None,
                    "contactType": None,
                }
            ],
            "nextPageLink": None,
            "count": 8,
        }

        m.get(self.van.connection.uri + "codes", json=json)
        assert_matching_tables(json["items"], self.van.get_codes())

    @requests_mock.Mocker()
    def test_get_code(self, m):

        json = {
            "codeId": 1004916,
            "parentCodeId": None,
            "name": "Data Entry",
            "description": "for test.",
            "codePath": "Data Entry",
            "createdByName": "",
            "dateCreated": "2018-07-13T15:16:00Z",
            "supportedEntities": None,
            "codeType": "Tag",
            "campaign": None,
            "contactType": None,
        }

        m.get(self.van.connection.uri + "codes/1004916", json=json)
        self.assertEqual(json, self.van.get_code(1004916))

    @requests_mock.Mocker()
    def test_get_code_types(self, m):

        json = ["Tag", "SourceCode"]
        m.get(self.van.connection.uri + "codeTypes", json=json)
        self.assertEqual(json, self.van.get_code_types())

    @requests_mock.Mocker()
    def test_create_code(self, m):

        m.post(self.van.connection.uri + "codes", json=1004960, status_code=201)

        # Test that it doesn't throw and error
        r = self.van.create_code(
            "Test Code",
            supported_entities=[
                {"name": "Events", "is_searchable": True, "is_applicable": True}
            ],
        )

        self.assertEqual(r, 1004960)

    @requests_mock.Mocker()
    def test_update_code(self, m):

        # Test a good input
        m.put(self.van.connection.uri + "codes/1004960", status_code=204)
        self.van.update_code(1004960, name="Test")

        # Test a bad input
        m.put(self.van.connection.uri + "codes/100496Q", status_code=404)
        self.assertRaises(HTTPError, self.van.update_code, "100496Q")

    @requests_mock.Mocker()
    def test_delete_code(self, m):

        # Test a good input
        m.delete(self.van.connection.uri + "codes/1004960", status_code=204)
        self.van.delete_code(1004960)

        # Test a bad input
        m.delete(self.van.connection.uri + "codes/100496Q", status_code=404)
        self.assertRaises(HTTPError, self.van.delete_code, "100496Q")

    @requests_mock.Mocker()
    def test_get_code_supported_entities(self, m):

        json = ["Contacts", "Events", "Locations"]
        m.get(self.van.connection.uri + "codes/supportedEntities", json=json)
        self.assertEqual(json, self.van.get_code_supported_entities())



import json
import os
import unittest
from unittest import mock
from parsons import ActionKit, Table

from test.utils import assert_matching_tables

ENV_PARAMETERS = {
    "ACTION_KIT_DOMAIN": "env_domain",
    "ACTION_KIT_USERNAME": "env_username",
    "ACTION_KIT_PASSWORD": "env_password",
}


class TestActionKit(unittest.TestCase):
    def setUp(self):
        self.actionkit = ActionKit(
            domain="domain.actionkit.com", username="user", password="password"
        )
        self.actionkit.conn = mock.MagicMock()

    def tearDown(self):
        pass

    @mock.patch.dict(os.environ, ENV_PARAMETERS)
    def test_from_environ(self):
        actionkit = ActionKit()
        self.assertEqual(actionkit.domain, "env_domain")
        self.assertEqual(actionkit.username, "env_username")
        self.assertEqual(actionkit.password, "env_password")

    def test_base_endpoint(self):
        # Test the endpoint
        url = self.actionkit._base_endpoint("user")
        self.assertEqual(url, "https://domain.actionkit.com/rest/v1/user/")

        url = self.actionkit._base_endpoint("user", 1234)
        self.assertEqual(url, "https://domain.actionkit.com/rest/v1/user/1234/")

        url = self.actionkit._base_endpoint("user", "1234")
        self.assertEqual(url, "https://domain.actionkit.com/rest/v1/user/1234/")

    def test_delete_actionfield(self):
        # Test delete actionfield

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=204)
        self.actionkit.conn = resp_mock

        self.actionkit.delete_actionfield(123)
        self.actionkit.conn.delete.assert_called_with(
            "https://domain.actionkit.com/rest/v1/actionfield/123/",
        )

    def test_get_user(self):
        # Test get user
        self.actionkit.get_user(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/user/123/", params=None
        )

    def test_get_user_fields(self):
        self.actionkit.get_user_fields()
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/user/schema/", params=None
        )

    def test_create_user(self):
        # Test create user

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_user(email="test")
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/user/",
            data=json.dumps({"email": "test"}),
        )

    def test_update_user(self):
        # Test update user

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=202)
        self.actionkit.conn = resp_mock

        self.actionkit.update_user(123, last_name="new name")
        self.actionkit.conn.patch.assert_called_with(
            "https://domain.actionkit.com/rest/v1/user/123/",
            data=json.dumps({"last_name": "new name"}),
        )

    def test_update_event(self):
        # Test update event

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=202)
        self.actionkit.conn = resp_mock
        self.actionkit.update_event(123, is_approved="test")
        self.actionkit.conn.patch.assert_called_with(
            "https://domain.actionkit.com/rest/v1/event/123/",
            data=json.dumps({"is_approved": "test"}),
        )

    def test_delete_user(self):
        # Test delete user

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=204)
        self.actionkit.conn = resp_mock

        self.actionkit.delete_user(123)
        self.actionkit.conn.delete.assert_called_with(
            "https://domain.actionkit.com/rest/v1/user/123/",
        )

    def test_get_campaign(self):
        # Test get campaign
        self.actionkit.get_campaign(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/campaign/123/", params=None
        )

    def test_create_campaign(self):
        # Test create campaign

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_campaign(name="new_campaign", field="field")
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/campaign/",
            data=json.dumps({"name": "new_campaign", "field": "field"}),
        )

    def test_get_event(self):
        # Test get event
        self.actionkit.get_event(1)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/event/1/", params=None
        )

    def test_get_events(self):
        # Test get events
        resp_mock = mock.MagicMock()
        type(resp_mock.get()).status_code = mock.PropertyMock(return_value=201)
        type(resp_mock.get()).json = lambda x: {"meta": {"next": ""}, "objects": []}
        self.actionkit.conn = resp_mock

        self.actionkit.get_events(100, order_by="created_at")
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/event/",
            params={"order_by": "created_at", "_limit": 100},
        )

    def test_get_event_create_page(self):
        # Test get event create page
        self.actionkit.get_event_create_page(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/eventcreatepage/123/", params=None
        )

    def test_create_event_create_page(self):
        # Test create event create page

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_event_create_page(
            name="new_page", campaign_id="123", title="title"
        )
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/eventcreatepage/",
            data=json.dumps(
                {
                    "campaign": "/rest/v1/campaign/123/",
                    "name": "new_page",
                    "title": "title",
                }
            ),
        )

    def test_get_event_create_form(self):
        # Test get event create form
        self.actionkit.get_event_create_form(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/eventcreateform/123/", params=None
        )

    def test_create_event_create_form(self):
        # Test event create form

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_event_create_form(
            page_id="123", thank_you_text="thank you"
        )
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/eventcreateform/",
            data=json.dumps(
                {"page": "/rest/v1/eventcreatepage/123/", "thank_you_text": "thank you"}
            ),
        )

    def test_get_event_signup_page(self):
        # Test get event signup page
        self.actionkit.get_event_signup_page(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/eventsignuppage/123/", params=None
        )

    def test_create_event_signup_page(self):
        # Test create event signup page

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_event_signup_page(
            name="new_name", campaign_id="123", title="title"
        )
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/eventsignuppage/",
            data=json.dumps(
                {
                    "campaign": "/rest/v1/campaign/123/",
                    "name": "new_name",
                    "title": "title",
                }
            ),
        )

    def test_get_event_signup_form(self):
        # Test get event signup form
        self.actionkit.get_event_signup_form(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/eventsignupform/123/", params=None
        )

    def test_create_event_signup_form(self):
        # Test create event signup form

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_event_signup_form(
            page_id="123", thank_you_text="thank you"
        )
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/eventsignupform/",
            data=json.dumps(
                {"page": "/rest/v1/page/123/", "thank_you_text": "thank you"}
            ),
        )

    def test_update_event_signup(self):
        # Test update event signup

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=202)
        self.actionkit.conn = resp_mock
        self.actionkit.update_event_signup(123, email="test")
        self.actionkit.conn.patch.assert_called_with(
            "https://domain.actionkit.com/rest/v1/eventsignup/123/",
            data=json.dumps({"email": "test"}),
        )

    def test_get_mailer(self):
        # Test get mailer
        self.actionkit.get_mailer(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/mailer/123/", params=None
        )

    def test_create_mailer(self):
        # Test create mailer

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_mailer(
            fromline="test <test@test.com>",
            subjects=["test1", "test2"],
            html="<p>test</p>",
        )
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/mailer/",
            data=json.dumps(
                {
                    "fromline": "test <test@test.com>",
                    "subjects": ["test1", "test2"],
                    "html": "<p>test</p>",
                }
            ),
        )

    def test_rebuild_mailer(self):
        # Test rebuild mailer

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.rebuild_mailer(123)
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/mailer/123/rebuild/",
            data=json.dumps({}),
        )

    def test_queue_mailer(self):
        # Test queue mailer

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.queue_mailer(123)
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/mailer/123/queue/",
            data=json.dumps({}),
        )

    def test_paginated_get(self):
        # Test paginated_get
        resp_mock = mock.MagicMock()
        first_mock = mock.MagicMock()
        second_mock = mock.MagicMock()
        first_mock.status_code = 201
        first_mock.json = lambda: {
            "meta": {"next": "/rest/v1/user/abc"},
            "objects": list(map(lambda x: {"value": x}, [*range(100)])),
        }
        second_mock.status_code = 201
        second_mock.json = lambda: {
            "meta": {"next": "/rest/v1/user/def"},
            "objects": list(map(lambda x: {"value": x}, [*range(100, 200)])),
        }
        resp_mock.get.side_effect = [first_mock, second_mock]
        self.actionkit.conn = resp_mock
        results = self.actionkit.paginated_get("user", 150, order_by="created_at")
        self.assertEqual(results.num_rows, 150)
        calls = [
            unittest.mock.call(
                "https://domain.actionkit.com/rest/v1/user/",
                params={"order_by": "created_at", "_limit": 100},
            ),
            unittest.mock.call("https://domain.actionkit.com/rest/v1/user/abc"),
        ]
        self.actionkit.conn.get.assert_has_calls(calls)

    def test_paginated_get_custom_limit(self):
        # Test paginated_get
        resp_mock = mock.MagicMock()
        first_mock = mock.MagicMock()
        second_mock = mock.MagicMock()
        first_mock.status_code = 201
        first_mock.json = lambda: {
            "meta": {"next": "/rest/v1/user/abc"},
            "objects": list(map(lambda x: {"value": x}, [*range(100)])),
        }
        second_mock.status_code = 201
        second_mock.json = lambda: {
            "meta": {"next": "/rest/v1/user/def"},
            "objects": list(map(lambda x: {"value": x}, [*range(100, 200)])),
        }
        resp_mock.get.side_effect = [first_mock, second_mock]
        self.actionkit.conn = resp_mock
        results = self.actionkit.paginated_get_custom_limit("user", 150, "value", 102)
        self.assertEqual(results.num_rows, 102)
        self.assertEqual(results.column_data("value")[0], 0)
        self.assertEqual(results.column_data("value")[-1], 101)
        calls = [
            unittest.mock.call(
                "https://domain.actionkit.com/rest/v1/user/",
                params={"order_by": "value", "_limit": 100},
            ),
            unittest.mock.call("https://domain.actionkit.com/rest/v1/user/abc"),
        ]
        self.actionkit.conn.get.assert_has_calls(calls)

    def test_get_order(self):
        # Test get order
        self.actionkit.get_order(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/order/123/", params=None
        )

    def test_update_order(self):
        # Test update order

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=202)
        self.actionkit.conn = resp_mock
        self.actionkit.update_order(123, account="test")
        self.actionkit.conn.patch.assert_called_with(
            "https://domain.actionkit.com/rest/v1/order/123/",
            data=json.dumps({"account": "test"}),
        )

    def test_get_orders(self):
        # Test get orders
        resp_mock = mock.MagicMock()
        type(resp_mock.get()).status_code = mock.PropertyMock(return_value=201)
        type(resp_mock.get()).json = lambda x: {"meta": {"next": ""}, "objects": []}
        self.actionkit.conn = resp_mock

        self.actionkit.get_orders(100, order_by="created_at")
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/order/",
            params={"order_by": "created_at", "_limit": 100},
        )

    def test_update_paymenttoken(self):
        # Test update payment token

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=202)
        self.actionkit.conn = resp_mock

        self.actionkit.update_paymenttoken(1, status="inactive")
        self.actionkit.conn.patch.assert_called_with(
            "https://domain.actionkit.com/rest/v1/paymenttoken/1/",
            data=json.dumps({"status": "inactive"}),
        )

    def test_get_page_followup(self):
        # Test get page followup
        self.actionkit.get_page_followup(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/pagefollowup/123/", params=None
        )

    def test_create_page_followup(self):
        # Test create page followup

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_page_followup(signup_page_id="123", url="url")
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/pagefollowup/",
            data=json.dumps({"page": "/rest/v1/eventsignuppage/123/", "url": "url"}),
        )

    def test_get_survey_question(self):
        # Test get survey question
        self.actionkit.get_survey_question(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/surveyquestion/123/", params=None
        )

    def test_update_survey_question(self):
        # Test update survey question

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=202)
        self.actionkit.conn = resp_mock
        self.actionkit.update_survey_question(123, question_html="test")
        self.actionkit.conn.patch.assert_called_with(
            "https://domain.actionkit.com/rest/v1/surveyquestion/123/",
            data=json.dumps({"question_html": "test"}),
        )

    def test_get_orderrecurring(self):
        # Test get orderrecurring
        self.actionkit.get_orderrecurring(123)
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/orderrecurring/123/", params=None
        )

    def test_cancel_orderrecurring(self):
        # Test cancel recurring order

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.cancel_orderrecurring(1)
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/orderrecurring/1/cancel/"
        )

    def test_update_orderrecurring(self):
        # Test update orderrecurring

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=202)
        self.actionkit.conn = resp_mock
        self.actionkit.update_orderrecurring(123, amount="1.00")
        self.actionkit.conn.patch.assert_called_with(
            "https://domain.actionkit.com/rest/v1/orderrecurring/123/",
            data=json.dumps({"amount": "1.00"}),
        )

    def test_create_transaction(self):
        # Test create transaction

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_transaction(
            account="Account",
            amount=1,
            amount_converted=1,
            currency="USD",
            failure_code="",
            failure_description="",
            failure_message="",
            order="/rest/v1/order/1/",
            status="completed",
            success=True,
            test_mode=False,
            trans_id="abc123",
            type="sale",
        )
        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/transaction/",
            data=json.dumps(
                {
                    "account": "Account",
                    "amount": 1,
                    "amount_converted": 1,
                    "currency": "USD",
                    "failure_code": "",
                    "failure_description": "",
                    "failure_message": "",
                    "order": "/rest/v1/order/1/",
                    "status": "completed",
                    "success": True,
                    "test_mode": False,
                    "trans_id": "abc123",
                    "type": "sale",
                }
            ),
        )

    def test_update_transaction(self):
        # Test update transaction

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=202)
        self.actionkit.conn = resp_mock
        self.actionkit.update_transaction(123, account="test")
        self.actionkit.conn.patch.assert_called_with(
            "https://domain.actionkit.com/rest/v1/transaction/123/",
            data=json.dumps({"account": "test"}),
        )

    def test_get_transactions(self):
        # Test get transactions
        resp_mock = mock.MagicMock()
        type(resp_mock.get()).status_code = mock.PropertyMock(return_value=201)
        type(resp_mock.get()).json = lambda x: {"meta": {"next": ""}, "objects": []}
        self.actionkit.conn = resp_mock

        self.actionkit.get_transactions(100, order_by="created_at")
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/transaction/",
            params={"order_by": "created_at", "_limit": 100},
        )

    def test_create_generic_action(self):
        # Test create a generic action

        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_generic_action(email="bob@bob.com", page="my_action")

        self.actionkit.conn.post.assert_called_with(
            "https://domain.actionkit.com/rest/v1/action/",
            data=json.dumps({"email": "bob@bob.com", "page": "my_action"}),
        )

    def test_bulk_upload_table(self):
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit._conn = lambda self: resp_mock
        self.actionkit.bulk_upload_table(
            Table(
                [
                    ("user_id", "user_customfield1", "action_foo"),
                    (5, "yes", "123 Main St"),
                ]
            ),
            "fake_page",
        )
        self.assertEqual(resp_mock.post.call_count, 2)
        name, args, kwargs = resp_mock.method_calls[1]
        self.assertEqual(
            kwargs["data"],
            {"page": "fake_page", "autocreate_user_fields": 0, "user_fields_only": 0},
        )
        upload_data = kwargs["files"]["upload"].read()
        self.assertEqual(
            upload_data.decode(),
            "user_id,user_customfield1,action_foo\r\n5,yes,123 Main St\r\n",
        )

    def test_bulk_upload_table_userfields(self):
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit._conn = lambda self: resp_mock
        self.actionkit.bulk_upload_table(
            Table([("user_id", "user_customfield1"), (5, "yes")]), "fake_page"
        )
        self.assertEqual(resp_mock.post.call_count, 2)
        name, args, kwargs = resp_mock.method_calls[1]
        self.assertEqual(
            kwargs["data"],
            {"page": "fake_page", "autocreate_user_fields": 0, "user_fields_only": 1},
        )
        self.assertEqual(
            kwargs["files"]["upload"].read().decode(),
            "user_id,user_customfield1\r\n5,yes\r\n",
        )

    def test_table_split(self):
        test1 = Table([("x", "y", "z"), ("a", "b", ""), ("1", "", "3"), ("4", "", "6")])
        tables = self.actionkit._split_tables_no_empties(test1, True, [])
        self.assertEqual(len(tables), 2)
        assert_matching_tables(tables[0], Table([("x", "y"), ("a", "b")]))
        assert_matching_tables(tables[1], Table([("x", "z"), ("1", "3"), ("4", "6")]))

        test2 = Table(
            [("x", "y", "z"), ("a", "b", "c"), ("1", "2", "3"), ("4", "5", "6")]
        )
        tables2 = self.actionkit._split_tables_no_empties(test2, True, [])
        self.assertEqual(len(tables2), 1)
        assert_matching_tables(tables2[0], test2)

        test3 = Table(
            [("x", "y", "z"), ("a", "b", ""), ("1", "2", "3"), ("4", "5", "6")]
        )
        tables3 = self.actionkit._split_tables_no_empties(test3, False, ["z"])
        self.assertEqual(len(tables3), 2)
        assert_matching_tables(tables3[0], Table([("x", "y"), ("a", "b")]))
        assert_matching_tables(
            tables3[1], Table([("x", "y", "z"), ("1", "2", "3"), ("4", "5", "6")])
        )

    def test_collect_errors(self):
        self.actionkit.collect_upload_errors([{"id": "12345"}])
        self.actionkit.conn.get.assert_called_with(
            "https://domain.actionkit.com/rest/v1/uploaderror/",
            params={"upload": "12345"},
        )



"""NGPVAN Locations Endpoints"""

from parsons.etl.table import Table
import logging

logger = logging.getLogger(__name__)


class Locations(object):
    def __init__(self, van_connection):

        self.connection = van_connection

    def get_locations(self, name=None):
        """
        Get locations.

        `Args:`
            name: str
                Filter locations by name.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request("locations", params={"name": name}))
        logger.info(f"Found {tbl.num_rows} locations.")
        return self._unpack_loc(tbl)

    def get_location(self, location_id):
        """
        Get a location.

        `Args:`
            location_id: int
                The location id.
        `Returns:`
            dict
        """

        r = self.connection.get_request(f"locations/{location_id}")
        logger.info(f"Found location {location_id}.")
        return r

    def create_location(
        self,
        name,
        address_line1=None,
        address_line2=None,
        city=None,
        state=None,
        zip_code=None,
    ):
        """
        Find or create a location. If location already exists, will return location id.

        `Args:`
            name: str
                A name for this location, no longer than 50 characters.
            address_line1: str
                First line of a street address.
            address_line2: str
                Second line of a street address.
            city: str
                City or town name.
            state: str
                Two or three character state or province code (e.g., MN, ON, NSW, etc.).
            zip_code: str
                ZIP, ZIP+4, Postal Code, Post code, etc.
            `Returns:`
                int
                    A location id.
        """

        location = {
            "name": name,
            "address": {
                "addressLine1": address_line1,
                "addressLine2": address_line2,
                "city": city,
                "stateOrProvince": state,
                "zipOrPostalCode": zip_code,
            },
        }

        r = self.connection.post_request("locations/findOrCreate", json=location)
        logger.info(f"Location {r} created.")
        return r

    def delete_location(self, location_id):
        """
        Delete a location.

        `Args:`
            location_id: int
                The location id
        `Returns:`
            ``None``
        """

        r = self.connection.delete_request(f"locations/{location_id}")
        logger.info(f"Location {location_id} deleted.")
        return r

    def _unpack_loc(self, table):
        # Internal method to unpack location json

        if isinstance(table, tuple):
            return table

        if "address" in table.columns:
            table.unpack_dict("address", prepend=False)

        if "geoLocation" in table.columns:
            table.unpack_dict("geoLocation", prepend=False)

        return table
